import os
from web3 import Web3
from .signer import Signer

class W3:
    def __init__(self, p, a=None):
        """
        Parameters:
            p (Web3.Provider) Provider being used to connect
            a (address) An optional default account to use. Will default to .eth.accounts[0] if omitted
        """

        self.instance = Web3(p)

        # normalize .account from the various scenarios
        if a != None:
            self.account = a
        else:
            if bool(self.instance.eth.default_account):
                self.account = self.instance.eth.default_account
            else:
                self.account = self.instance.eth.accounts[0]

        self.signer = Signer()

    def contract(self, address, abi):
        """Get an instance of the vedor low-level contract object

        Parameters:
            address (address) The address of the deployed smart contract
            abi (string) ABI of the deployed smart contract

        Returns:
            the vendor specific contract object
        """
        
        return self.instance.eth.contract(address=address, abi=abi)

    def sign_order(self, o, i, a):
        """Sign an order, producing an EIP712 compliant signature

        Parameters:
            o (dict) Swivel Order object
            i (int) ChainId
            a (string) Address of the deployed verifying contract

        Returns:
            The signature hex
        """

        key = os.getenv('PRIVATE_KEY')
        return self.signer.sign_order(o, i, a, self.instance.toBytes(hexstr=key))

    def estimate_gas(self, a):
        """Return an estimate of gas to be used in a transaction, as well as the price of the gas
        
        Parameters:
            a (tuple) Length 2 tuple whose values are a web3 transactable object and a tx_opts dict.
            This is the common return of all Swivel.py H.O.C contract methods. 

        Returns:
            gas, gas_price
        """

        built = self.build_transaction(a)
        return built['gas'], built['gasPrice']

    def build_transaction(self, a):
        """Return a suitable transaction object for signing
        
        Description:
            As per our pattern a[1] is the tx_opts dict. Note that you should include, at least,
            'chainId' in that dict. Other optional properties are available, see
            https://web3py.readthedocs.io/en/latest/web3.eth.account.html#sign-a-transaction.
            * 'maxFeePerGas'
            * 'maxPriorityFeePerGas'
            * etc...
        """

        a[1]['nonce'] = self.instance.eth.get_transaction_count(a[1]['from'])
        return a[0].buildTransaction(a[1])

    def sign_transaction(self, t):
        """Called after a transaction has been built with `build_transaction`
        
        Parameters:
            t (dict) A built transaction object from build_transaction

        Returns:
            A signed transaction
        """

        key = os.getenv('PRIVATE_KEY')
        return self.instance.eth.account.sign_transaction(t, private_key=key)

    def send_raw_transaction(self, t):
        """Given a raw signed transaction, broadcast it
        
        Parameters:
            t (transaction) A web3 transaction which has been signed

        Returns:
            A transaction hash suitable for web3's ...wait_for_transaction_receipt method
        """

        return self.instance.eth.send_raw_transaction(t.rawTransaction)

    def send(self, a):
        """Convenience method which builds, signs and broadcasts a transaction"""

        return self.send_raw_transaction(self.sign_transaction(self.build_transaction(a)))



from abc import abstractmethod
from swivel.abstracts import Deployed

class MarketPlace(Deployed):
    @abstractmethod
    def admin(self, opts=None):
        """The stored admin address for this contract

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def transfer_admin(self, a, opts=None):
        """Allows the current admin to transfer the title to another party

        Description:
            Note that this method is only callable by the admin

        Parameters:
            a (address) Address of the new admin

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def swivel(self, opts=None):
        """The address of the associated Swivel contract deployment

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def set_swivel_address(self, a, opts=None):
        """Sets the address of the Swivel smart contract this market place is associated with

        Description:
            Note that this method is only callable by the admin once

        Parameters:
            a (address) The address of a deployed Swivel smart contract
            opts (dict) Optional transaction options

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def pause(self, b, opts=None):
        """Allows the admin to pause / unpause market transactions

        Parameters:
            b Boolean which acts as a toggle, True to pause, False to unpause

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def paused(self, opts=None):
        """A boolean which indicates a pause in all markets if truthy

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def c_token_address(self, u, m, opts=None):
        """Gets the compound token address associated with a given market

        Parameters:
            u (string) Underlying token address
            m (int) Maturity epoch
            opts (dict) optional call opts

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def create_market(self, m, c, n, s, opts=None):
        """Creates a new market

        Description:
            New instances of both ZcToken and VaultTracker are deployed, their addresses then being associated
            with the newly created market.

            Note that this method is only callable by the admin

        Parameters:
            m (int) Epoch in seconds, the maturity of the market
            c (string) Address of the Compound token associated with the market
            n (string) Name for the new ZcToken
            s (string) Name for the new ZcToken
            opts (dict) Optional transaction options

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def markets(self, u, m, opts=None):
        """Gets the market associated with the given underlying and maturity arguments

        Description:
            The returned Market object is { cTokenAddr, ZcTokenAddr, vaultAddr }

        Parameters:
            u (string) Underlying token address
            m (int) Muturity epoch
            opts (dict) optional call opts

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def mature_market(self, u, m, opts=None):
        """Called after maturity, allowing all of the zcTokens to earn floating interest on Compound until funds are released

        Parameters:
            u (string) Underlying token address
            m (int) Maturity epoch
            opts (dict) optional call opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def transfer_vault_notional(self, u, m, t, a, opts=None):
        """Transfer vault notional from sender to a given address

        Parameters:
            u (string) Underlying token address
            m (int) Muturity epoch
            t (string) Address of the amount owner
            a (int) Amount to transfer
            opts (dict) Optional transaction opts

        Returns:
            web3 transactable, opts
        """

        pass



from py_eth_sig_utils.signing import sign_typed_data, v_r_s_to_signature
from swivel.constants import DOMAIN_NAME, DOMAIN_VERSION, HEX_PREFIX

class Signer:
    def sign_order(self, o, i, a, k):
        """Sign an order, producing an EIP712 compliant signature

        Parameters:
            o (dict) Swivel Order object
            i (int) ChainId
            a (string) Address of the deployed verifying contract
            k (bytes) Private key as bytes

        Returns:
            The signature hex with 0x prefix normalized
        """
        
        data = self.prepare_data()
        self.prepare_message(data, o)
        self.prepare_domain(data, i, a)

        sig = v_r_s_to_signature(*sign_typed_data(data, k)).hex()

        if sig.startswith(HEX_PREFIX):
            return sig
        else:
            return HEX_PREFIX + sig

    def prepare_message(self, d, o):
        """Update the data dictionary with the 'message' entry (Order)"""

        msg = { 'message': o }

        d.update(msg)

    def prepare_domain(self, d, i, a):
        """Update the data dictionary with the 'domain' entry"""

        domain = {
            'domain': {
                'name': DOMAIN_NAME,
                'version': DOMAIN_VERSION,
                'chainId': i,
                'verifyingContract': a,
            }
        }

        d.update(domain)

    def prepare_data(self):
        """Get the initial EIP712 compliant data dictionary to sign"""

        return {
            'types': {
                'EIP712Domain': [
                    { 'name': 'name', 'type': 'string' },
                    { 'name': 'version', 'type': 'string' },
                    { 'name': 'chainId', 'type': 'uint256' },
                    { 'name': 'verifyingContract', 'type': 'address' },
                ],
                'Order': [
                    { 'name': 'key', 'type': 'bytes32' },
                    { 'name': 'maker', 'type': 'address' },
                    { 'name': 'underlying', 'type': 'address' },
                    { 'name': 'vault', 'type': 'bool' },
                    { 'name': 'exit', 'type': 'bool' },
                    { 'name': 'principal', 'type': 'uint256' },
                    { 'name': 'premium', 'type': 'uint256' },
                    { 'name': 'maturity', 'type': 'uint256' },
                    { 'name': 'expiry', 'type': 'uint256' },
                ]
            },
            'primaryType': 'Order'
        }



import pytest
from swivel.helpers import call

@pytest.fixture(scope='module')
def key(vendor):
    return vendor.instance.toBytes(hexstr='0xfb28b03032bbb105e1199e496b23a6435a077375cbea9c6c4998b971a672873c')

@pytest.fixture(scope='module')
def orders(key):
    # we can use the same order 2x here for the purpose of this test...
    order = {
        'key': key,
        'maker': '0x7111F9Aeb2C1b9344EC274780dc9e3806bdc60Ef',
        'underlying': '0x5592EC0cfb4dbc12D3aB100b257153436a1f0FEa',
        'vault': False,
        'exit': False,
        'principal': 1000,
        'premium': 60,
        'maturity': 1655255622,
        'expiry': 1625173101,
    }

    return (order, order)

@pytest.fixture(scope='module')
def signatures(vendor):
    # again, just using same thing 2x
    sig = '06aa9b426a48ff92d3b940eea498ba8bf9cf4e3e05987bdf4c68c66382385f5e59dd5f013c2b13eb7958234ad441c53675b196d88d143eccf93192b5b9758d571b'
    return (sig, sig)

@pytest.fixture(scope='module')
def amounts():
    return (1000, 500)

def test_name(swivel):
    name = call(swivel.name())
    assert name == 'Swivel Finance'

def test_version(swivel):
    verz = call(swivel.version())
    assert verz == '2.0.0'

def test_hold(swivel):
    hold = call(swivel.hold())
    assert hold == 259200

def test_domain(swivel):
    dom = call(swivel.domain())
    assert dom != None

def test_market_place_address(market_place, swivel):
    m_place_addr = call(swivel.market_place())
    assert m_place_addr == market_place.address

def test_swivel_admin(swivel):
    addr = call(swivel.admin())
    assert addr == swivel.vendor.account

def test_transfer_admin(swivel, orders):
    txable, opts = swivel.transfer_admin(orders[0]['maker'])

    assert callable(txable)
    assert isinstance(opts, dict)

def test_feenominators(swivel):
    fee1 = call(swivel.feenominators(0))
    assert fee1 == 200
    fee2 = call(swivel.feenominators(1))
    assert fee2 == 600
    fee3 = call(swivel.feenominators(2))
    assert fee3 == 400
    fee4 = call(swivel.feenominators(3))
    assert fee4 == 200

def test_min_feenominator(swivel):
    the_min = call(swivel.min_feenominator())
    assert the_min == 33

def test_initiate(swivel, orders, amounts, signatures):
    txable, opts = swivel.initiate(orders, amounts, signatures, opts={ 'gas': 1000000 })

    # if the call encoded correctly we'll get back a callable and a dict
    assert callable(txable)
    assert isinstance(opts, dict)

def test_exit(swivel, orders, amounts, signatures):
    txable, opts = swivel.exit(orders, amounts, signatures, opts={ 'gas': 250000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_cancel(swivel, orders, signatures):
    txable, opts = swivel.cancel(orders[0], signatures[0], opts={ 'gas': 50000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_split_underlying(swivel):
    txable, opts = swivel.split_underlying('0x5592ec0cfb4dbc12d3ab100b257153436a1f0fea',
        1234567890, 1000, opts={ 'gas': 17000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_combine_tokens(swivel):
    txable, opts = swivel.combine_tokens('0x5592ec0cfb4dbc12d3ab100b257153436a1f0fea', 1234567890, 500, opts={ 'gas': 1000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_redeem_zctoken(swivel):
    txable, opts = swivel.redeem_zc_token('0x5592ec0cfb4dbc12d3ab100b257153436a1f0fea', 1234567890, 700, opts={ 'gas': 2000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_redeem_vault_interest(swivel):
    txable, opts = swivel.redeem_vault_interest('0x5592ec0cfb4dbc12d3ab100b257153436a1f0fea', 1234567899, opts={ 'gas': 2500 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_redeem_swivel_vault_interest(swivel):
    txable, opts = swivel.redeem_vault_interest('0x5592ec0cfb4dbc12d3ab100b257153436a1f0fea', 1234567999, opts={ 'gas': 2000 })

    assert callable(txable)
    assert isinstance(opts, dict)



from abc import ABC

class Deployed(ABC):
    def at(self, a, opts=None):
        """Get a reference to the vendor specific contract instance deployed at a given address

        Parameters:
            a (string) the address of the deployed smart contract
            o (dict) optional tx options
        """

        self.address = a
        # TODO throw if the contract is falsy?
        self.contract = self.vendor.contract(a, self.abi)
        # TODO do we want to keep this reference?
        self.opts = opts

    def tx_opts(self, o):
        """Sets default transaction options"""

        defaults = { 'from': self.vendor.account }

        if o is not None:
            # any user passed opts should overwrite defaults...
            defaults.update(o)

        return defaults



from swivel.abstracts import Deployed
from .abi import CERC20

class CErc20(Deployed):
    def __init__(self, v):
        self.vendor = v
        self.abi = CERC20

    def exchange_rate_current_returns(self, n, opts=None):
        return self.contract.functions.exchangeRateCurrentReturns(n), self.tx_opts(opts)



import pytest

def test_account(vendor):
    assert len(vendor.account) == 42
    # the fixture does not pass an account, so it should be accounts[0]
    assert vendor.account == vendor.instance.eth.accounts[0]

def test_estimate_gas(vendor, swivel):
    args = swivel.split_underlying('0x5592EC0cfb4dbc12D3aB100b257153436a1f0FEa',
        1234567890, 1000, opts={ 'gas': 25000, 'chainId': 1 })

    gas, price = vendor.estimate_gas(args)

    assert gas != None
    assert price != None

def test_build_transaction(vendor, swivel):
    # build up the tx, opts args first
    args = swivel.split_underlying('0x5592EC0cfb4dbc12D3aB100b257153436a1f0FEa',
        1234567890, 1000, opts={ 'gas': 25000, 'chainId': 1 })
    
    built = vendor.build_transaction(args)
    # as long as 'data' is present we'll assume success
    assert len(built['data']) > 0

def test_sign_transaction(vendor, swivel):
    # we need to include from in this one or the random accounts[0] will cause it to fail...
    args = swivel.split_underlying('0x5592EC0cfb4dbc12D3aB100b257153436a1f0FEa',
        1234567890, 1000, opts={ 'from': '0x7111F9Aeb2C1b9344EC274780dc9e3806bdc60Ef', 'gas': 25000, 'chainId': 1 })
    
    built = vendor.build_transaction(args)
    signed = vendor.sign_transaction(built)

    assert signed != None



class Caller:
    """A class used to mock the `.call` able objects returned from Web3 calls"""
    def call(o=None):
        pass

class Transactor:
    """A class used to mock the `.transact` able objects returned from Web3 transaction calls"""
    def transact(o):
        pass




import pytest
from swivel.helpers import call, transact

@pytest.fixture(scope='module')
def swivel_address():
    return '0x8e7bFA3106c0544b6468833c0EB41c350b50A5CA'

@pytest.fixture(scope='module')
def underlying(market_place):
    return market_place.vendor.instance.toChecksumAddress('0x5592ec0cfb4dbc12d3ab100b257153436a1f0fea')

@pytest.fixture(scope='module')
def maturity():
    return 1633988168

@pytest.fixture(scope='module')
def c_token(market_place):
    return market_place.vendor.instance.toChecksumAddress('0x6d7f0754ffeb405d23c51ce938289d4835be3b14')

def test_m_place_admin(market_place):
    addr = call(market_place.admin())
    assert addr == market_place.vendor.account

def test_m_place_transfer_admin(market_place, underlying):
    # the address doesn't matter, just checking that the abi accepts the method...
    txable, opts = market_place.transfer_admin(underlying)

    assert callable(txable)
    assert isinstance(opts, dict)

def test_swivel(market_place, swivel_address):
    #set it first...
    tx_hash = transact(market_place.set_swivel_address(swivel_address))
    tx_rcpt = market_place.vendor.instance.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_rcpt != None

    addr = call(market_place.swivel())
    assert addr == swivel_address

def test_paused(market_place):
    paused = call(market_place.paused())
    assert paused == False

def test_pause(market_place):
    txable, opts = market_place.pause(True)

    assert callable(txable)
    assert isinstance(opts, dict)

def test_create_market(market_place, underlying, maturity, c_token):
    name = 'token'
    sym = 'tkn'
    txable, opts = market_place.create_market(maturity, c_token, name, sym, opts={ 'gas': 50000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_c_token_addr(market_place, underlying, maturity, c_token):
    clable, _ = market_place.c_token_address(underlying, maturity)

    assert callable(clable)

def test_mature_market(market_place, underlying, maturity):
    txable, opts = market_place.mature_market(underlying, maturity, { 'gas': 100000 })

    assert callable(txable)
    assert isinstance(opts, dict)

def test_transfer_vault_notional(market_place, underlying, maturity):
    txable, opts = market_place.transfer_vault_notional(underlying, maturity, '0x50MeDud3', 1000, { 'gas': 200000 })

    assert callable(txable)
    assert isinstance(opts, dict)



from swivel.abstracts import MarketPlace as base
from .abi import MARKET_PLACE

class MarketPlace(base):
    def __init__(self, v):
        """
        Parameters:
            v (W3) Instance of a vendor W3 class (no other vendors are supported as of now)
        """
        self.vendor = v
        self.abi = MARKET_PLACE

    def admin(self, opts=None):
        return self.contract.functions.admin(), opts

    def transfer_admin(self, a, opts=None):
        return self.contract.functions.transferAdmin(a), self.tx_opts(opts)

    def swivel(self, opts=None):
        return self.contract.functions.swivel(), opts

    def set_swivel_address(self, a, opts=None):
        return self.contract.functions.setSwivelAddress(a), self.tx_opts(opts)

    def pause(self, b, opts=None):
        return self.contract.functions.pause(b), self.tx_opts(opts)

    def paused(self, opts=None):
        return self.contract.functions.paused(), opts

    def c_token_address(self, u, m, opts=None):
        return self.contract.functions.cTokenAddress(u, m), opts

    def create_market(self, m, c, n, s, opts=None):
        return self.contract.functions.createMarket(m, c, n, s,), self.tx_opts(opts)

    def markets(self, u, m, opts=None):
        return self.contract.functions.markets(u, m), opts

    def mature_market(self, u, m, opts=None):
        return self.contract.functions.matureMarket(u, m), self.tx_opts(opts)

    def transfer_vault_notional(self, u, m, t, a, opts=None):
        return self.contract.functions.transferVaultNotional(u, m, t, a), self.tx_opts(opts)



import os
import pytest
import logging
from web3 import Web3, EthereumTesterProvider
from eth_tester import PyEVMBackend, EthereumTester
from swivel.vendors import W3
from swivel.contracts.bin import MARKET_PLACE, SWIVEL
from swivel.helpers import transact
from swivel.contracts import MarketPlace, Swivel

@pytest.fixture(scope='module')
def logger():
    return logging.getLogger(__name__)

@pytest.fixture(scope='module')
def eth_tester():
    override = {'gas_limit': 6700000}
    params = PyEVMBackend._generate_genesis_params(overrides=override)
    backend = PyEVMBackend(params)
    return EthereumTester(backend)

@pytest.fixture(scope='module')
def provider(eth_tester):
    return EthereumTesterProvider(eth_tester)

@pytest.fixture(scope='module')
def vendor(provider):
    return W3(provider)

@pytest.fixture(scope='module')
def market_place(vendor):
    # the HOC has it's abi available
    m_place = MarketPlace(vendor)
    # our W3 vendor .contract method is not for deployment, use the primitive here
    deployed = vendor.instance.eth.contract(abi=m_place.abi, bytecode=MARKET_PLACE)
    # market place needs much gas to deploy...
    tx_opts = { 'from': vendor.account, 'gas': 6000000 }
    tx_hash = deployed.constructor().transact(tx_opts)
    tx_rcpt = vendor.instance.eth.wait_for_transaction_receipt(tx_hash)
    m_place.at(tx_rcpt['contractAddress'])
    # NOTE that the swivel address is not set as of yet...
    return m_place

@pytest.fixture(scope='module')
def swivel(vendor, market_place):
    swiv = Swivel(vendor)
    deployed = vendor.instance.eth.contract(abi=swiv.abi, bytecode=SWIVEL)
    # likely does not need 6M, but doesn't matter...
    tx_opts = {'from': vendor.account, 'gas': 6000000 }
    # a verifier address is passed to the swivel ctor now - this one is the DEV rinkeby deploy
    tx_hash = deployed.constructor(market_place.address, '0x4ccD4C002216f08218EdE1B13621faa80CecfC98').transact(tx_opts)
    tx_rcpt = vendor.instance.eth.wait_for_transaction_receipt(tx_hash)
    swiv.at(tx_rcpt['contractAddress'])
    return swiv

@pytest.fixture(scope='module')
def market_place_with_swivel(market_place, swivel):
    tx_hash = transact(market_place.set_swivel_address(swivel.address))
    vendor.instance.eth.wait_for_transaction_receipt(tx_hash)
    return market_place



from swivel.abstracts.deployed import Deployed
from swivel.abstracts.vault_tracker import VaultTracker
from swivel.abstracts.market_place import MarketPlace
from swivel.abstracts.swivel import Swivel



from swivel.abstracts import VaultTracker as base
from .abi import VAULT_TRACKER

class VaultTracker(base):
    def __init__(self, v):
        """
        Parameters:
            v (W3) Instance of a vendor W3 class (no other vendors are supported as of now)
        """
        self.vendor = v
        self.abi = VAULT_TRACKER

    def admin(self, opts=None):
        return self.contract.functions.admin(), opts

    def c_token_address(self, opts=None):
        return self.contract.functions.cTokenAddr(), opts

    def swivel(self, opts=None):
        return self.contract.functions.swivel(), opts

    def maturity(self, opts=None):
        return self.contract.functions.maturity(), opts

    def maturity_rate(self, opts=None):
        return self.contract.functions.maturityRate(), opts

    def vaults(self, o, opts=None):
        return self.contract.functions.vaults(o), opts
    
    def balances_of(self, o, opts=None):
        return self.contract.functions.balancesOf(o), opts



from abc import abstractmethod
from swivel.abstracts import Deployed

class Swivel(Deployed):
    @abstractmethod
    def name(self, opts=None):
        """The stored name constant for this contract

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def version(self, opts=None):
        """The stored version constant for this contract

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def hold(self, opts=None):
        """The stored hold constant for this contract

        Description:
            TODO

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def min_feenominator(self, opts=None):
        """The stored constant representing the minimum value for any fee denominator

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def domain(self, opts=None):
        """The stored EIP712 domain hash for this contract

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def market_place(self, opts=None):
        """The stored address of a market place contract associated with this contract

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def admin(self, opts=None):
        """The stored admin address for this contract

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def transfer_admin(self, a, opts=None):
        """Allows the current admin to transfer the title to another party

        Description:
            Note that this method is only callable by the admin

        Parameters:
            a (address) Address of the new admin

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def feenominators(self, i, opts=None):
        """The stored fee constants for this contract

        Parameters:
            i (int) Index of the fenominator array to return

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def initiate(self, orders, a, s, opts=None):
        """Allows a user to initiate a position

        Parameters:
            orders (tuple) Offline swivel orders
            a (list) Order volume (principal) amounts relative to orders list
            s (tuple) Valid ECDSA signatures for each order
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def exit(self, orders, a, s, opts=None):
        """Allows a user to exit (sell) a currently held position to the marketplace

        Parameters:
            orders (tuple) Offline swivel orders
            a (list) Order volume (principal) amounts relative to orders list
            s (tuple) Valid ECDSA signatures for each order
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def cancel(self, order, s, opts=None):
        """Allows the cancellation of an order, preventing it from being filled further

        Parameters:
            order (dict) An offline swivel order
            s (string) Valid ECDSA signature for the order
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass
        
    @abstractmethod
    def split_underlying(self, u, m, a, opts=None):
        """Allows users to deposit underlying and in the process split it into/mint zcTokens and vault notional

        Parameters:
            u (string) Address of the underlying token
            m (int) Maturity timestamp of the market
            a (int) Amount of underlying being deposited
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def combine_tokens(self, u, m, a, opts=None):
        """Allows users to deposit/burn 1:1 amounts of both zcTokens and vault notional, in the process 'combining' the two and redeeming underlying

        Parameters:
            u (string) Address of the underlying token
            m (int) Maturity timestamp of the market
            a (int) Amount of zctokens being redeemed
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def redeem_zc_token(self, u, m, a, opts=None):
        """Allows zctoken holders to redeem their tokens for underlying tokens after maturity has been reached

        Parameters:
            u (string) Address of the underlying token
            m (int) Maturity timestamp of the market
            a (int) Amount of zctokens being redeemed
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def redeem_vault_interest(self, u, m, opts=None):
        """Allows vault owners to redeem any currently accrued interest

        Parameters:
            u (string) Address of the underlying token
            m (int) Maturity timestamp of the market
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass

    @abstractmethod
    def redeem_swivel_vault_interest(self, u, m, opts=None):
        """Allows Swivel redeem any currently accrued interest

        Parameters:
            u (string) Address of the underlying token
            m (int) Maturity timestamp of the market
            opts (dict) Optional tx opts

        Returns:
            web3 transactable, opts
        """

        pass



CERC20 = '[{"inputs":[],"name":"exchangeRateCurrent","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"exchangeRateCurrentReturns","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"mint","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"mintCalled","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"mintReturns","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"redeemUnderlying","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"redeemUnderlyingCalled","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint256","name":"n","type":"uint256"}],"name":"redeemUnderlyingReturns","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"underlying","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"a","type":"address"}],"name":"underlyingReturns","outputs":[],"stateMutability":"nonpayable","type":"function"}]'



import pytest
from swivel.contracts.bin import VAULT_TRACKER
from swivel.helpers import call
from swivel.contracts import VaultTracker
from swivel.mocks.bin import CERC20
from swivel.mocks import CErc20

@pytest.fixture(scope='module')
def c_token(vendor):
    cerc20 = CErc20(vendor)
    deployed = vendor.instance.eth.contract(abi=cerc20.abi, bytecode=CERC20)
    tx_hash = deployed.constructor().transact({ 'from': vendor.account })
    tx_rcpt = vendor.instance.eth.wait_for_transaction_receipt(tx_hash)
    cerc20.at(tx_rcpt['contractAddress'])
    return cerc20

@pytest.fixture(scope='module')
def vault_tracker(vendor, c_token):
    # the HOC has it's abi available
    tracker = VaultTracker(vendor)
    # our W3 vendor .contract method is not for deployment, use the primitive here
    deployed = vendor.instance.eth.contract(abi=tracker.abi, bytecode=VAULT_TRACKER)

    # web3 will complain about a non-checksum address like the ctoken...
    # c_token_addr = vendor.instance.toChecksumAddress('0x6d7f0754ffeb405d23c51ce938289d4835be3b14')

    tx_hash = deployed.constructor(123456789, c_token.address, '0x8e7bFA3106c0544b6468833c0EB41c350b50A5CA').transact({ 'from': vendor.account })
    tx_rcpt = vendor.instance.eth.wait_for_transaction_receipt(tx_hash)
    tracker.at(tx_rcpt['contractAddress'])
    return tracker

@pytest.fixture
def vault():
    return { 'notional': 1000, 'redeemable': 500, 'exchangeRate': 10 }

def test_admin(vault_tracker):
    addr = call(vault_tracker.admin())
    # the vendor will normalize .account...
    assert addr == vault_tracker.vendor.account

def test_c_token_addr(vault_tracker, c_token):
    addr = call(vault_tracker.c_token_address())
    # assert addr == vault_tracker.vendor.instance.toChecksumAddress('0x6d7f0754ffeb405d23c51ce938289d4835be3b14')
    assert addr == c_token.address

def test_swivel_addr(vault_tracker):
    addr = call(vault_tracker.swivel())
    assert addr == '0x8e7bFA3106c0544b6468833c0EB41c350b50A5CA'

def test_maturity(vault_tracker):
    mty = call(vault_tracker.maturity())
    assert mty == 123456789

def test_maturity_rate(vault_tracker):
    rate = call(vault_tracker.maturity_rate())
    assert rate == 0

def test_vaults(vault_tracker, vault):
    called, opts = vault_tracker.vaults('0xG1mM3mYVaU1t')

    assert callable(called)
    assert opts == None

def test_balances_of(vault_tracker, vault):
    called, opts = vault_tracker.balances_of('0xG1mM3mYBa1anC35')

    assert callable(called)
    assert opts == None



from setuptools import setup, find_packages

with open('README.md', 'r') as file:
    long_description = file.read()

test_deps = [
    'pytest==6.2.5',
    'eth-tester==0.5.0b4',
    'web3==5.23.1',
    'py-eth-sig-utils==0.4.0',
]

extras = {
    'test': test_deps,
}

setup(
    name='swivel',
    version='2.2.0',
    author='Swivel Finance',
    author_email='support@swivel.finance',
    description='A python library for interacting with the Swivel Finance Protocol',
    long_description=long_description,
    keywords='swivel, finance, ethereum, blockchain, defi, web3',
    long_description_content_type='text/markdown',
    url='https://github.com/Swivel-Finance/swivel-py',
    packages=find_packages(exclude='tests'),
    python_requires='>=3.7.3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    tests_require=test_deps,
    extras_require=extras,
)



import pytest
from py_eth_sig_utils.signing import recover_typed_data, signature_to_v_r_s
from swivel.constants import DOMAIN_NAME, DOMAIN_VERSION, HEX_PREFIX

@pytest.fixture(scope='module')
def key(vendor):
    return vendor.instance.toBytes(hexstr='0xfb28b03032bbb105e1199e496b23a6435a077375cbea9c6c4998b971a672873c')

@pytest.fixture(scope='module')
def order(key):
    return {
        'key': key,
        'maker': '0x7111F9Aeb2C1b9344EC274780dc9e3806bdc60Ef',
        'underlying': '0x5592EC0cfb4dbc12D3aB100b257153436a1f0FEa',
        'vault': False,
        'exit': False,
        'principal': 1000,
        'premium': 60,
        'maturity': 1655255622,
        'expiry': 1625173101,
    }

def test_prepare_data(vendor):
    data = vendor.signer.prepare_data()
    assert len(data['types']['EIP712Domain']) == 4
    assert len(data['types']['Order']) == 9
    assert data['primaryType'] == 'Order'

def test_prepare_domain(vendor):
    data = vendor.signer.prepare_data()
    vendor.signer.prepare_domain(data, 4, '0xSp1M3gG5')
    assert data['domain']['name'] == DOMAIN_NAME
    assert data['domain']['version'] == DOMAIN_VERSION
    assert data['domain']['chainId'] == 4
    assert data['domain']['verifyingContract'] == '0xSp1M3gG5'
    # peek at the others to see if they are there...
    assert len(data['types']['Order']) == 9

def test_prepare_message(vendor):
    data = vendor.signer.prepare_data()
    key = '0xK3y123'
    principal = '1000000000000000'
    order = { 'key': key, 'vault': True, 'exit': False, 'principal': principal }
    vendor.signer.prepare_message(data, order)
    assert data['message']['key'] == key
    assert data['message']['vault'] == True
    assert data['message']['exit'] == False
    assert data['message']['principal'] == principal
    # peek at the others to see if they are there...
    assert len(data['types']['EIP712Domain']) == 4

def test_sign_order(vendor, order):
    verifier = '0x25b71690A99A692707f6F4933A76a58ECDD0b9Ac'
    sig = vendor.sign_order(order, 4, verifier)
    assert len(sig) > 0
    assert sig.startswith(HEX_PREFIX)

    # recover it
    data = vendor.signer.prepare_data()
    vendor.signer.prepare_message(data, order)
    vendor.signer.prepare_domain(data, 4, verifier)
    sig_bytes = vendor.instance.toBytes(hexstr=sig)
    address = recover_typed_data(data, *signature_to_v_r_s(sig_bytes))
    assert address == order['maker']




from py_eth_sig_utils.signing import signature_to_v_r_s
from swivel.helpers import v_r_s_to_dict
from swivel.abstracts import Swivel as base
from .abi import SWIVEL

class Swivel(base):
    def __init__(self, v):
        """
        Parameters:
            v (W3) Instance of a vendor W3 class (no other vendors are supported as of now)
        """
        self.vendor = v
        self.abi = SWIVEL

    def name(self, opts=None):
        return self.contract.functions.NAME(), opts

    def version(self, opts=None):
        return self.contract.functions.VERSION(), opts

    def hold(self, opts=None):
        return self.contract.functions.HOLD(), opts

    def domain(self, opts=None):
        return self.contract.functions.domain(), opts

    def market_place(self, opts=None):
        return self.contract.functions.marketPlace(), opts

    def admin(self, opts=None):
        return self.contract.functions.admin(), opts

    def transfer_admin(self, a, opts=None):
        return self.contract.functions.transferAdmin(a), self.tx_opts(opts)

    def min_feenominator(self, opts=None):
        return self.contract.functions.MIN_FEENOMINATOR(), opts

    def feenominators(self, i, opts=None):
        return self.contract.functions.feenominators(i), opts

    def initiate(self, orders, a, s, opts=None):
        # normalize the full signatures to a tuple of vrs components (tuples)
        components = tuple(map(lambda sig: v_r_s_to_dict(signature_to_v_r_s(self.vendor.instance.toBytes(hexstr=sig))), s))
        return self.contract.functions.initiate(orders, a, components), self.tx_opts(opts)

    def exit(self, orders, a, s, opts=None):
        # normalize the full signatures to a list of vrs components
        components = tuple(map(lambda sig: v_r_s_to_dict(signature_to_v_r_s(self.vendor.instance.toBytes(hexstr=sig))), s))
        return self.contract.functions.exit(orders, a, components), self.tx_opts(opts)

    def cancel(self, order, s, opts=None):
        components = signature_to_v_r_s(self.vendor.instance.toBytes(hexstr=s))
        return self.contract.functions.cancel(order, v_r_s_to_dict(components)), self.tx_opts(opts)

    def split_underlying(self, u, m, a, opts=None):
        return self.contract.functions.splitUnderlying(u, m, a), self.tx_opts(opts)

    def combine_tokens(self, u, m, a, opts=None):
        return self.contract.functions.combineTokens(u, m, a), self.tx_opts(opts)
    
    def redeem_zc_token(self, u, m, a, opts=None):
        return self.contract.functions.redeemZcToken(u, m, a), self.tx_opts(opts)

    def redeem_vault_interest(self, u, m, opts=None):
        return self.contract.functions.redeemVaultInterest(u, m), self.tx_opts(opts)

    def redeem_swivel_vault_interest(self, u, m, opts=None):
        return self.contract.functions.redeemSwivelVaultInterest(u, m), self.tx_opts(opts)



from abc import abstractmethod
from swivel.abstracts import Deployed

class VaultTracker(Deployed):
    """Persists and curates Vault objects associated with users

    Note that a Vault is a Dict object in this form:

        { notional, redeemable, exchangerate }
    """

    @abstractmethod
    def admin(self, opts=None):
        """The stored admin address for this contract

        Note that this should always be the MarketPlace address

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def c_token_address(self, opts=None):
        """Adress of the compound token referenced by this vault tracker

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def swivel(self, opts=None):
        """Address of the associated Swivel contract deployment

        Returns:
            web3 callable, opts
        """

        pass
    
    @abstractmethod
    def maturity(self, opts=None):
        """An Epoch (in seconds) representing the time of maturity

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def maturity_rate(self, opts=None):
        """The maturity rate (TODO: better description)

        Returns:
            web3 callable, opts
        """

        pass

    @abstractmethod
    def vaults(self, o, opts=None):
        """Get a Vault for a given address

        Parameters:
            o (string) Address of the vault owner

        Returns:
            web3 callable, opts
        """

        pass
    
    @abstractmethod
    def balances_of(self, o, opts=None):
        """Get Vault balances for a given user

        Parameters:
            o (string) Address of the vault owner

        Returns:
            web3 callable, opts
        """
        pass



from web3 import Web3

def call(args):
    return args[0].call(args[1])

def transact(args):
    """For unlocked accounts, we can simply call transact directly"""

    return args[0].transact(args[1])

# TODO **?
def v_r_s_to_dict(t):
    """Transform the py-eth-sig-utils VRS tuple into a dict"""

    return {'v': t[0], 'r': Web3.toBytes(t[1]), 's': Web3.toBytes(t[2])}
